from django.conf import settings

from django_jwt.roles import ROLE

OIDC_AUDIENCE = getattr(settings, "OIDC_AUDIENCE", ["account", "broker"])

# key from KeyCloak; value is user model
OIDC_USER_UPDATE = getattr(settings, "OIDC_USER_UPDATE", True)
OIDC_USER_MODIFIED_FIELD = getattr(settings, "OIDC_USER_MODIFIED_FIELD", "modified_timestamp")
OIDC_TOKEN_MODIFIED_FIELD = getattr(settings, "OIDC_TOKEN_MODIFIED_FIELD", "updated_at")
OIDC_TOKEN_USER_UID = getattr(settings, "OIDC_TOKEN_USER_UID", "sub")
OIDC_USER_UID = getattr(settings, "OIDC_USER_UID", "kc_id")
OIDC_USER_MAPPING = getattr(
    settings,
    "OIDC_USER_MAPPING",
    {
        "given_name": "first_name",
        "family_name": "last_name",
        "name": "username",
    },
)
OIDC_USER_DEFAULTS = getattr(
    settings,
    "OIDC_USER_DEFAULTS",
    {
        "is_active": True,
    },
)
OIDC_USER_ON_CREATE = getattr(
    settings,
    "OIDC_USER_ON_CREATE",
    None,
)
OIDC_USER_ON_UPDATE = getattr(
    settings,
    "OIDC_USER_ON_UPDATE",
    None,
)

OIDC_CONFIG_ROUTES = getattr(settings, "OIDC_CONFIG_ROUTES", {})
OIDC_ADMIN_ISSUER = getattr(settings, "OIDC_ADMIN_ISSUER", None)
OIDC_ADMIN_CLIENT_ID = getattr(settings, "OIDC_ADMIN_CLIENT_ID", "cs-completeanatomy-admin")
OIDC_ADMIN_SCOPE = getattr(settings, "OIDC_ADMIN_SCOPE", "openid")
OIDC_ADMIN_ROLES = getattr(settings, "OIDC_ADMIN_ROLES", [])
OIDC_AUTHORIZATION_BACKEND = getattr(
    settings, "OIDC_AUTHORIZATION_BACKEND", "django.contrib.auth.backends.ModelBackend",
)

if not OIDC_ADMIN_ISSUER:
    OIDC_ADMIN_ISSUER = OIDC_CONFIG_ROUTES.get("ES256", None)

for role in OIDC_ADMIN_ROLES:
    assert isinstance(role, ROLE), f"Role must be a namedtuple, got {type(role)}"

assert isinstance(OIDC_USER_MAPPING, dict) or callable(
    OIDC_USER_MAPPING
), "OIDC_USER_MAPPING must be a dict or function"



from datetime import datetime
from functools import cache
from logging import getLogger

import pytz
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.http.request import HttpRequest

from django_jwt import settings
from django_jwt.utils import oidc_handler

utc = pytz.UTC
log = getLogger(__name__)

model = get_user_model()


def mapper(user_data: dict) -> dict:
    if callable(settings.OIDC_USER_MAPPING):
        return settings.OIDC_USER_MAPPING(user_data)
    return {ca_key: user_data[kc_key] for kc_key, ca_key in settings.OIDC_USER_MAPPING.items() if kc_key in user_data}


class UserHandler:
    modified_at = None

    def __init__(self, payload: dict, request: HttpRequest, access_token: str):
        # payload of access token without user info
        self.kwargs = settings.OIDC_USER_DEFAULTS.copy()
        self.kwargs[settings.OIDC_USER_UID] = payload[settings.OIDC_TOKEN_USER_UID]

        modified_at = payload.get(settings.OIDC_TOKEN_MODIFIED_FIELD, None)
        if modified_at and isinstance(modified_at, int):
            self.modified_at = utc.localize(datetime.fromtimestamp(modified_at))

        self.on_create = settings.OIDC_USER_ON_CREATE
        self.on_update = settings.OIDC_USER_ON_UPDATE
        self.request = request
        self.access_token = access_token

    def _collect_user_data(self):
        """Collect user data from KeyCloak"""

        user_data = oidc_handler.get_user_info(self.access_token)
        log.info(f"User data: {self.kwargs}, access_token: {self.access_token}")
        self.kwargs["email"] = user_data["email"].lower()
        self.kwargs.update(mapper(user_data))

    def _update_user(self, user):
        """Update user fields if they are changed"""

        self._collect_user_data()
        if hasattr(user, settings.OIDC_USER_MODIFIED_FIELD):
            self.kwargs[settings.OIDC_USER_MODIFIED_FIELD] = self.modified_at
        for key, val in self.kwargs.items():
            if getattr(user, key) != val:
                setattr(user, key, val)
        user.save(update_fields=self.kwargs.keys())

    def _create_new_user(self) -> model:
        """Create new user if user is not found in database even by email."""

        user = model.objects.create(**self.kwargs)
        if self.on_create:
            self.on_create(user, self.request)
        return user

    def _get_by_email(self) -> model:
        """Get user from database by email and update to resave kc_id."""

        user = model.objects.get(email=self.kwargs["email"])
        self._update_user(user)
        if self.on_update:
            self.on_update(user, self.request)
        return user

    def _update_existing_user(self, user):
        """Check modified_at and update user if it is changed in KeyCloak."""

        user_modified_at = getattr(user, settings.OIDC_USER_MODIFIED_FIELD, None)
        if user_modified_at:
            if not user_modified_at.tzinfo:
                user_modified_at = utc.localize(user_modified_at)
            if self.modified_at and user_modified_at < self.modified_at:
                self._update_user(user)
                if self.on_update:
                    self.on_update(user, self.request)

    def get_user(self) -> model:
        """
        Get user from database by kc_id or email.
        If user is not found, create new user.
        Update user fields if they are changed in KeyCloak.
        """

        try:
            user = model.objects.get(**{settings.OIDC_USER_UID: self.kwargs[settings.OIDC_USER_UID]})
            self._update_existing_user(user)
            return user

        except model.DoesNotExist:
            self._collect_user_data()
            try:
                return self._get_by_email()
            except model.DoesNotExist:
                return self._create_new_user()


class RoleHandler:
    """
    Process user roles and permissions from access token.
    Token be like:
    ...
    "resource_access": {
        "complete_anatomy": {
            "roles": [
                "admin"
            ]
        }
    },
    """

    @property
    def roles(self) -> dict:
        return {role.name: role for role in settings.OIDC_ADMIN_ROLES}

    @cache
    def get_permissions(self, role_name: str) -> Permission:
        return Permission.objects.filter(codename__in=self.roles[role_name].permissions)

    @cache
    def get_groups(self, role_name: str) -> Group:
        return Group.objects.filter(name__in=self.roles[role_name].groups)

    def apply(self, user: model, access_token: dict) -> list[str]:
        token_roles = access_token.get("resource_access", {}).get(settings.OIDC_ADMIN_CLIENT_ID, {}).get("roles", [])
        for role_name in token_roles:
            if role_name in self.roles:
                role = self.roles[role_name]
                user.groups.add(*self.get_groups(role_name))
                user.user_permissions.add(*self.get_permissions(role_name))
                user.is_staff = True
                if role.is_superuser != user.is_superuser:
                    user.is_superuser = role.is_superuser
                user.save(update_fields=["is_superuser", "is_staff"])
                break

        return token_roles


role_handler = RoleHandler()



import base64
import random
import string
from logging import getLogger
from urllib.parse import urlencode

from django.conf import settings
from django.contrib.auth import login
from django.core.cache import cache
from django.http.response import HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views import View
from requests.exceptions import HTTPError

from django_jwt import settings as jwt_settings
from django_jwt.config import config
from django_jwt.exceptions import BadRequestException, ConfigException
from django_jwt.pkce import PKCESecret
from django_jwt.user import UserHandler, role_handler
from django_jwt.utils import get_access_token, oidc_handler

log = getLogger(__name__)


def silent_sso_check(request):
    return HttpResponse("<html><body><script>parent.postMessage(location.href, location.origin)</script></body></html>")


def index_response(request, msg, status=400):
    logout_url = config.admin().get("end_session_endpoint")
    return render(
        request,
        "django-jwt-index.html",
        {
            "error_message": msg,
            "login_url": reverse("start_oidc_auth"),
            "logout_url": logout_url,
            "redirect_uri": request.build_absolute_uri(reverse("start_oidc_auth")),
        },
        status=status,
    )


class AbsView(View):
    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except HTTPError as exc:
            log.warning(f"OIDC Admin HTTPError: {exc}")
            return index_response(request=request, msg=exc.response.text, status=exc.response.status_code)
        except ConfigException as exc:
            return HttpResponse(content=str(exc), status=500)
        except BadRequestException as exc:
            return index_response(request=request, msg=str(exc))
        except Exception as exc:
            return index_response(request=request, msg=str(exc))


class StartOIDCAuthView(AbsView):
    def get(self, request):
        pkce_secret = PKCESecret()
        redirect_uri = request.build_absolute_uri(reverse("receive_redirect_view"))
        authorization_endpoint = config.admin().get("authorization_endpoint")
        state = base64.urlsafe_b64encode(
            "".join(random.choices(string.ascii_letters + string.digits + "-._~", k=32)).encode()
        ).decode()
        random_nonce = "".join(random.choices(string.ascii_letters + string.digits + "-._~", k=32))
        params = {
            "client_id": jwt_settings.OIDC_ADMIN_CLIENT_ID,
            "redirect_uri": redirect_uri,
            "response_type": "code",
            "state": state,
            "scope": jwt_settings.OIDC_ADMIN_SCOPE,
            "code_challenge": pkce_secret.challenge,
            "code_challenge_method": pkce_secret.challenge_method,
            "ui_locales": "en",
            "nonce": random_nonce,
        }
        cache.set(state, str(pkce_secret), timeout=600)
        log.info(f"OIDC Admin login: {authorization_endpoint}?{urlencode(params)}")
        return redirect(f"{authorization_endpoint}?{urlencode(params)}")


class ReceiveRedirectView(AbsView):
    def get(self, request):
        code = request.GET.get("code")
        state = request.GET.get("state")
        if not code or not state:
            log.warning(f"No code or state in the request {request.GET}")
            raise BadRequestException("No code or state in the request")

        redirect_uri = request.build_absolute_uri(reverse("receive_redirect_view"))
        if state := cache.get(state):
            token = get_access_token(code, redirect_uri, state)
            data = oidc_handler.decode_token(token)
            user = UserHandler(data, request, token).get_user()
            log.info(f"OIDC Admin login: {user}", extra={"data": data})
            roles = role_handler.apply(user, data)
            if not user.is_staff:
                raise BadRequestException(f"User {user.email} is not staff\nRoles: {roles}")
            login(request, user, backend=jwt_settings.OIDC_AUTHORIZATION_BACKEND)
            return redirect("admin:index")

        raise BadRequestException("No PKCE secret found in cache")


class LogoutView(AbsView):
    def get(self, request):
        return index_response(request, "Logged out", status=401)



import base64
import hashlib
import random
import string


class PKCESecret:
    """PKCE secret."""

    def __init__(self, length: int = 128):
        self.value = "".join(random.choices(string.ascii_letters + string.digits, k=length))

    def __str__(self) -> str:
        return self.value

    def __bytes__(self) -> bytes:
        return self.value.encode()

    @property
    def challenge(self) -> bytes:
        """PKCE challenge matching the secret value."""
        return base64.urlsafe_b64encode(hashlib.sha256(bytes(self)).digest()).rstrip(b"=")

    @property
    def challenge_method(self) -> str:
        """PKCE challenge method, always 'S256' in this implementation."""
        return "S256"



from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils import timezone


class User(AbstractUser):
    kc_id = models.CharField(max_length=255, unique=True, null=True, blank=True)
    modified_timestamp = models.DateTimeField(auto_now=False, default=timezone.now)
